@@ -12,10 +12,11 @@
  *	void S_INIT_LOCK(slock_t *lock)
  *		Initialize a spinlock (to the unlocked state).
  *
- *	void S_LOCK(slock_t *lock)
+ *	int S_LOCK(slock_t *lock)
  *		Acquire a spinlock, waiting if necessary.
  *		Time out and abort() if unable to acquire the lock in a
  *		"reasonable" amount of time --- typically ~ 1 minute.
+ *		Should return number of "delays"; see s_lock.c
  *
  *	void S_UNLOCK(slock_t *lock)
  *		Unlock a previously acquired lock.
@@ -83,7 +84,7 @@
  *	when using the SysV semaphore code.
  *
  *
- * Portions Copyright (c) 1996-2012, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2013, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *	  src/include/storage/s_lock.h
@@ -335,6 +336,29 @@ tas(volatile slock_t *lock)
 #endif	 /* __arm__ */
 
 
+/*
+ * On ARM64, we use __sync_lock_test_and_set(int *, int) if available.
+ */
+#if defined(__aarch64__) || defined(__aarch64)
+#ifdef HAVE_GCC_INT_ATOMICS
+#define HAS_TEST_AND_SET
+
+#define TAS(lock) tas(lock)
+
+typedef int slock_t;
+
+static __inline__ int
+tas(volatile slock_t *lock)
+{
+	return __sync_lock_test_and_set(lock, 1);
+}
+
+#define S_UNLOCK(lock) __sync_lock_release(lock)
+
+#endif	 /* HAVE_GCC_INT_ATOMICS */
+#endif	 /* __aarch64__ */
+
+
 /* S/390 and S/390x Linux (32- and 64-bit zSeries) */
 #if defined(__s390__) || defined(__s390x__)
 #define HAS_TEST_AND_SET
@@ -461,8 +485,7 @@ do \
 
 
 /* Linux Motorola 68k */
-#if (defined(__mc68000__) || defined(__m68k__)) && \
-    (defined(__linux__) || defined(__OpenBSD__))
+#if (defined(__mc68000__) || defined(__m68k__)) && defined(__linux__)
 #define HAS_TEST_AND_SET
 
 typedef unsigned char slock_t;
@@ -680,41 +703,39 @@ tas(volatile slock_t *lock)
 #endif	 /* __sh__ */
 
 
-/* Motorola 88k */
-#if defined(__m88k__)
+/* These live in s_lock.c, but only for gcc */
+
+
+#if defined(__m68k__) && !defined(__linux__)	/* non-Linux Motorola 68k */
 #define HAS_TEST_AND_SET
 
 typedef unsigned char slock_t;
+#endif
+
+
+#if defined(__m88k__)		/* Motorola 88k */
+#define HAS_TEST_AND_SET
+
+typedef unsigned int slock_t;
 
 #define TAS(lock) tas(lock)
 
 static __inline__ int
 tas(volatile slock_t *lock)
 {
-	register slock_t rv = 1;
+	register slock_t _res = 1;
 
-	__asm__	__volatile__(
-		"	xmem.bu	%0, %2, %%r0	\n"
-:		"+r"(rv), "+m"(*lock)
-:		"r"(lock)
+	__asm__ __volatile__(
+		"	xmem	%0, %2, %%r0\n"
+:		"=r"(_res)
+:		"0" (_res), "r"(lock)
 :		"memory");
-
-	return (int) rv;
+	return (int) _res;
 }
 
 #endif	 /* __m88k__ */
 
 
-/* These live in s_lock.c, but only for gcc */
-
-
-#if defined(__m68k__) && !(defined(__linux__) || defined(__OpenBSD__))	/* non-Linux Motorola 68k */
-#define HAS_TEST_AND_SET
-
-typedef unsigned char slock_t;
-#endif
-
-
 #endif	/* defined(__GNUC__) || defined(__INTEL_COMPILER) */
 
 
@@ -1004,10 +1025,7 @@ extern int	tas_sema(volatile slock_t *lock);
 
 #if !defined(S_LOCK)
 #define S_LOCK(lock) \
-	do { \
-		if (TAS(lock)) \
-			s_lock((lock), __FILE__, __LINE__); \
-	} while (0)
+	(TAS(lock) ? s_lock((lock), __FILE__, __LINE__) : 0)
 #endif	 /* S_LOCK */
 
 #if !defined(S_LOCK_FREE)
@@ -1041,7 +1059,7 @@ extern int	tas(volatile slock_t *lock);		/* in port/.../tas.s, or
 /*
  * Platform-independent out-of-line support routines
  */
-extern void s_lock(volatile slock_t *lock, const char *file, int line);
+extern int s_lock(volatile slock_t *lock, const char *file, int line);
 
 /* Support for dynamic adjustment of spins_per_delay */
 #define DEFAULT_SPINS_PER_DELAY  100
